==> delegate6.1.18.tar.gz: delegate6.1.18/src/ccache.c <==
/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1997 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	ccache.c (Connection Cache)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	970125	created
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include "vsignal.h"
#include "delegate.h"
#include <errno.h>
int MAX_CC = 16;

extern FILE *fopenCC();
extern FILE *fcloseCC();

#define LOCALHOST	"127.0.0.1"

static tryOpen(Conn,user,new,path,portp)
	Connection *Conn;
	char *user;
	char *path;
	int *portp;
{	FILE *ccfp;
	char url[1024],server[1024],buff[128];
	int cci,svport,svpid,sock;
	int rcode;

	/* ServerSiceCC should not connect to another ServerSideCC
	 * of the same protocol with itself...
	 */
	if( !new && ImCC /*&& strcmp(DST_PROTO,ImCC_PROTO) == 0*/ ){
		sv1log("serverCC: don't connect svCC from svCC\n");
		return -1;
	}

	sock = -1;
	ccfp = NULL;

	if( user[0] == 0 )
		user = "anonymous";
	sprintf(url,"%s-%s-%s-%d",DST_PROTO,user,DST_HOST,DST_PORT);

	for( cci = 0; cci < MAX_CC; cci++ ){
		sprintf(server,"%s-%d",url,cci);
		if( ccfp = fopenCC(server,"r+",path) ){
			if( new ){
				fcloseCC();
				continue;
			}
			if( lock_exclusiveNB(fileno(ccfp)) != 0 ){
				fcloseCC();
				continue;
			}
			svport = -1;
			buff[0] = 0;
			fgets(buff,sizeof(buff),ccfp);
			sscanf(buff,"%d %d",&svport,&svpid);
			if( svpid == getpid() ){
				sv1log("serverCC: don't connect to myself\n");
				fcloseCC();
				continue;
			}

			sock = client_open("serverCC",DST_PROTO,LOCALHOST,svport);
			fcloseCC();

			if( 0 <= sock ){
				*portp = svport;
				break;
			}
			sv1log("serverCC: salvaged [%s] %s\n",path,buff);
			unlink(path);

		}else{
			if( !new )
				break;

			if( (ccfp = fopenCC(server,"w+",path)) == NULL )
				continue;

			if( sock == -1 )
				sock = server_open("serverCC",LOCALHOST,0,1);
			svport = sockPort(sock);
			sprintf(buff,"%d %d\n",svport,getpid());
			fputs(buff,ccfp);
			fcloseCC();

			*portp = svport;
			break;
		}
	}
	if( 0 < sock )
		sv1log("serverCC: %s [%s] %s",new?"wrote":"read",path,buff);
	else if( new )
		sv1log("serverCC: cannot open Connection Cache file\n");
	return sock;
}

extern char *TIMEFORM_ANSI_C;

static char mypath[256];
static void sigTERM(sig)
{
	sv1log("SIGTERM for BoundProxy: %s\n",mypath);
	unlink(mypath);
	Finish(0);
}

beBoundProxy(Conn,user,timeout,func,a1,a2,a3,a4)
	Connection *Conn;
	char *user;
	int (*func)();
	void *a1,*a2,*a3,*a4;
{	int svsock,svport,clsock,clport;
	char path[1024],stime[128],clhost[256];
	int rcode;
	FILE *ccfp;
	int ccc,start,done;
	int startSerno;
	int fdv[3],rfdv[3];

	if( ServViaCc )
		return;

	checkCloseOnTimeout(0);
	closedups(0);

	if( Conn->xf_filters ){
		/* no good on Windows */
		sv1log("#### NO CC with CFI filters = 0x%x\n",Conn->xf_filters);
		return;
	}

	svsock = tryOpen(Conn,user,1,path,&svport);
	if( svsock < 0 )
		return;

	sayStartCC();

	strcpy(mypath,path);
	Vsignal(SIGINT,sigTERM);
	Vsignal(SIGTERM,sigTERM);
	close_FSERVER(Conn,0);

	if( ccfp = fopen(path,"a") )
	for( ccc = 1; ; ccc++ ){
		clear_DGclnt(Conn);
		ProcTitle(Conn,"(standby=%s://%s:%d)",DST_PROTO,DST_HOST,DST_PORT);

		fdv[0] = svsock;
		fdv[1] = ToS;
		fdv[2] = FromS;

		if( PollIns(timeout*1000,3,fdv,rfdv) <= 0 ){
			sv1log("serverCC: exit by timeout=%ds\n",timeout);
			break;
		}
		if( !IsConnected(ToS,NULL) || !IsConnected(FromS,NULL) ){
			sv1log("serverCC: exit by stale CC [%d=%d][%d=%d].\n",
				ToS,rfdv[1],FromS,rfdv[2]);
			break;
		}
		if( PollIn(FromS,1) != 0 ){
			sv1log("serverCC: exit by something from server.\n");
			break;
		}
		clsock = ACCEPT(svsock,1,-1,10);
		if( clsock < 0 ){
			sv1log("serverCC: exit by accept() failure\n");
			break;
		}

/*
BIND/CONNECT AT LOCLHOST PORT ONLY
		if( !localsocket(clsock) ){
			clport = getpeerNAME(clsock,clhost);
			sv1log("#### forbidden access from remote host(%s)\n",
				clhost,clport);
			close(clsock);
			continue;
		}
*/

		rcode = lock_exclusiveTO(fileno(ccfp),1000,NULL);
		start = time(0L);
		startSerno = RequestSerno;
		StrftimeLocal(stime,sizeof(stime),TIMEFORM_ANSI_C,start,0);
		fprintf(ccfp,"[%s] start\n",stime);
		fflush(ccfp);
		sv1log("%sCC [%d] restart\n",DST_PROTO,ccc);

		ProcTitle(Conn,"(%s://%s:%d)",DST_PROTO,DST_HOST,DST_PORT);
		FromC = ToC = ClientSock = clsock;
		ImCC = 1;

		setsockbuf(clsock,32*1024,32*1024);
		rcode = (*func)(Conn,a1,a2,a3,a4);
		ImCC = 0;
		close(clsock);

		done = time(0L);
		StrftimeLocal(stime,sizeof(stime),TIMEFORM_ANSI_C,done,0);
		fprintf(ccfp,"[%s] done, %d seconds\n",stime,done-start);
		fflush(ccfp);
		lock_unlock(fileno(ccfp));
		sv1log("%sCC [%d] done, %d requests / %d seconds\n",
			DST_PROTO,ccc,RequestSerno-startSerno,done-start);

		if( rcode != 0 ){
			sv1log("serverCC: exit by exit code=%d\n",rcode);
			break;
		}
	}
	close(svsock);
	fclose(ccfp);
	unlink(path);
	sayFinishCC();
}

connectToCache(Conn,user,svsockp)
	Connection *Conn;
	char *user;
	int *svsockp;
{	char path[1024];
	int clsock,svport;

	clsock = tryOpen(Conn,user,0,path,&svport);
	if( 0 <= clsock ){
		*svsockp = clsock;
		ServViaCc = 1;
		return 1;
	}else	return 0;
}


static int CC_master;
static int CC_Report[2] = {-1,-1};
static int CC_Active = 0;
static int CC_Procs[256];

set_CC()
{
	if( CC_Report[0] < 0 ){
		CC_master = getpid();
		Socketpair(CC_Report);
		setNonblockingIO(CC_Report[1],1);
		setCloseOnExec(CC_Report[0]);
		setCloseOnExec(CC_Report[1]);
	}
}
putCCinfo(pid)
{	int fd;

	if( (fd = CC_Report[1]) < 0 )
		return;
	write(fd,&pid,sizeof(pid));
}
sayStartCC(){
	/* When the CC_master is not the parent of this process,
	 * it will not able to count down this CC by wait().
	 * Such CC will be fork()ed as a NNTPCC on timeout...
	 */
	if( getppid() != CC_master )
		return;
	putCCinfo(getpid());
}
sayFinishCC(){
	if( getppid() != CC_master )
		return;
	putCCinfo(-getpid());

}
num_CC()
{	int fd,pid;

	if( (fd = CC_Report[0]) < 0 )
		return 0;

	if( getpid() != CC_master )
		return 0;

	while( readTimeoutBlocked(fd,&pid,sizeof(pid),1) == sizeof(pid) ){
		sv1log("## got_CCinfo: %d\n",pid);
		if( 0 <= pid ){
			sv1log("## add_CC [%d] %d\n",CC_Active,pid);
			CC_Procs[CC_Active++] = pid;
		}else{
			del_CC(-pid,"byRead");
		}
	}
	return CC_Active;
}
del_CC(pid,how)
	char *how;
{	int ci;

	for( ci = 0; ci < CC_Active; ci++ ){
		if( CC_Procs[ci] == pid ){
			sv1log("## del_CC [%d/%d] %d %s\n",
				ci,CC_Active,pid,how);
			for(; ci < CC_Active; ci++ )
				CC_Procs[ci] = CC_Procs[ci+1];
			CC_Active--;
			break;
		}
	}
}
kill_CC()
{	int ci,pid,nkill;

	nkill = 0;
	for( ci = 0; ci < CC_Active; ci++ ){
		if( 0 < (pid = CC_Procs[ci]) ){
			if( Kill(pid,SIGTERM) == 0 ){
				sv1log("Killed CC [pid=%d]\n",pid);
				nkill++;
			}
		}
	}
	return nkill;
}

==> delegate9.0.3-pre10.tar.gz: delegate9.0.3-pre10/src/ccache.c <==
/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1997 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	ccache.c (Connection Cache)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	970125	created
//////////////////////////////////////////////////////////////////////#*/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "ystring.h"
#include "vsignal.h"
#include "delegate.h"
#include "fpoll.h"
#include "proc.h"
#include "file.h"
int MAX_CC = 16;

void clear_DGclnt(Connection *Conn);
FILE *fopenCC(PCStr(server),PCStr(mode),PVStr(path));
FILE *fcloseCC();
void del_CC(int pid,PCStr(how));
void sayStartCC();
void sayFinishCC();

typedef struct {
	MStr(	ce_Mypath,256);
	int	ce_CC_master;
	int	ce_CC_Report[2];
	int	ce_CC_Active;
	int	ce_CC_Procs[256];
} CCacheEnv;
static CCacheEnv *ccacheEnv;
#define Mypath		ccacheEnv[0].ce_Mypath
/**/
#define CC_master	ccacheEnv[0].ce_CC_master
#define CC_Report	ccacheEnv[0].ce_CC_Report
#define CC_Active	ccacheEnv[0].ce_CC_Active
#define CC_Procs	ccacheEnv[0].ce_CC_Procs
void minit_ccache()
{
	if( ccacheEnv == 0 ){
		ccacheEnv = NewStruct(CCacheEnv);
		CC_Report[0] = -1;
		CC_Report[1] = -1;
	}
}


#define LOCALHOST	"127.0.0.1"

static int tryOpen(Connection *Conn,PCStr(user),int xnew,PVStr(path),int *portp)
{	FILE *ccfp;
	CStr(url,1024);
	CStr(server,1024);
	CStr(buff,128);
	int cci,svport,svpid,sock;
	int rcode;

	/* ServerSiceCC should not connect to another ServerSideCC
	 * of the same protocol with itself...
	 */
	if( !xnew && ImCC /*&& strcmp(DST_PROTO,ImCC_PROTO) == 0*/ ){
		sv1log("serverCC: don't connect svCC from svCC\n");
		return -1;
	}

	sock = -1;
	ccfp = NULL;

	if( user[0] == 0 )
		user = "anonymous";
	sprintf(url,"%s-%s-%s-%d",DST_PROTO,user,DST_HOST,DST_PORT);

	for( cci = 0; cci < MAX_CC; cci++ ){
		sprintf(server,"%s-%d",url,cci);
		if( ccfp = fopenCC(server,"r+",AVStr(path)) ){
			if( xnew ){
				fcloseCC();
				continue;
			}
			if( lock_exclusiveNB(fileno(ccfp)) != 0 ){
				fcloseCC();
				continue;
			}
			svport = -1;
			buff[0] = 0;
			fgets(buff,sizeof(buff),ccfp);
			sscanf(buff,"%d %d",&svport,&svpid);
			if( svpid == getpid() ){
				sv1log("serverCC: don't connect to myself\n");
				fcloseCC();
				continue;
			}

			sock = client_open("serverCC",DST_PROTO,LOCALHOST,svport);
			fcloseCC();

			if( 0 <= sock ){
				*portp = svport;
				break;
			}
			sv1log("serverCC: salvaged [%s] %s\n",path,buff);
			unlink(path);

		}else{
			if( !xnew )
				break;

			if( (ccfp = fopenCC(server,"w+",AVStr(path))) == NULL )
				continue;

			if( sock == -1 )
				sock = server_open("serverCC",CVStr(LOCALHOST),0,1);
			svport = sockPort(sock);
			sprintf(buff,"%d %d\n",svport,getpid());
			fputs(buff,ccfp);
			fcloseCC();

			*portp = svport;
			break;
		}
	}
	if( 0 < sock )
		sv1log("serverCC: %s [%s] %s",xnew?"wrote":"read",path,buff);
	else if( xnew )
		sv1log("serverCC: cannot open Connection Cache file\n");
	return sock;
}

static void sigTERM(int sig)
{
	sv1log("SIGTERM for BoundProxy: %s\n",Mypath);
	unlink(Mypath);
	Finish(0);
}

void beBoundProxy(Connection *Conn,PCStr(user),int timeout,iFUNCP func,...)
{	int svsock,svport,clsock,clport;
	CStr(path,1024);
	CStr(stime,128);
	CStr(clhost,256);
	int rcode;
	FILE *ccfp;
	int ccc,start,done;
	int startSerno;
	int fdv[3],rfdv[3];
	VARGS(8,func);

	if( ServViaCc )
		return;

	checkCloseOnTimeout(0);
	closedups(0);

	if( Conn->xf_filters ){
		/* no good on Windows */
		sv1log("#### NO CC with CFI filters = 0x%x\n",Conn->xf_filters);
		return;
	}

	svsock = tryOpen(Conn,user,1,AVStr(path),&svport);
	if( svsock < 0 )
		return;

	sayStartCC();

	strcpy(Mypath,path);
	Vsignal(SIGINT,sigTERM);
	Vsignal(SIGTERM,sigTERM);
	close_FSERVER(Conn,0);

	if( ccfp = fopen(path,"a") )
	for( ccc = 1; ; ccc++ ){
		clear_DGclnt(Conn);
		ProcTitle(Conn,"(standby=%s://%s:%d)",DST_PROTO,DST_HOST,DST_PORT);

		fdv[0] = svsock;
		fdv[1] = ToS;
		fdv[2] = FromS;

		if( PollIns(timeout*1000,3,fdv,rfdv) <= 0 ){
			sv1log("serverCC: exit by timeout=%ds\n",timeout);
			break;
		}
		if( !IsConnected(ToS,NULL) || !IsConnected(FromS,NULL) ){
			sv1log("serverCC: exit by stale CC [%d=%d][%d=%d].\n",
				ToS,rfdv[1],FromS,rfdv[2]);
			break;
		}
		if( PollIn(FromS,1) != 0 ){
			CStr(buf,128);
			int cc;
			cc = readTIMEOUT(FromS,AVStr(buf),sizeof(buf)-1);
			if( 0 < cc )
				buf[cc] = 0;
			else	buf[0] = 0;
			sv1log("serverCC: exit by something from server: %s\n",
				buf);
			break;
		}
		clsock = ACCEPT(svsock,1,-1,10);
		if( clsock < 0 ){
			sv1log("serverCC: exit by accept() failure\n");
			break;
		}

/*
BIND/CONNECT AT LOCLHOST PORT ONLY
		if( !localsocket(clsock) ){
			clport = getpeerNAME(clsock,clhost);
			sv1log("#### forbidden access from remote host(%s)\n",
				clhost,clport);
			close(clsock);
			continue;
		}
*/

		rcode = lock_exclusiveTO(fileno(ccfp),1000,NULL);
		start = time(0L);
		startSerno = RequestSerno;
		StrftimeLocal(AVStr(stime),sizeof(stime),TIMEFORM_ANSI_C,start,0);
		fprintf(ccfp,"[%s] start\n",stime);
		fflush(ccfp);
		sv1log("%sCC [%d] restart\n",DST_PROTO,ccc);

		ProcTitle(Conn,"(%s://%s:%d)",DST_PROTO,DST_HOST,DST_PORT);
		FromC = ToC = ClientSock = clsock;
		ImCC = 1;

		setsockbuf(clsock,32*1024,32*1024);
		rcode = (*func)(Conn,VA8);
		ImCC = 0;
		close(clsock);

		done = time(0L);
		StrftimeLocal(AVStr(stime),sizeof(stime),TIMEFORM_ANSI_C,done,0);
		fprintf(ccfp,"[%s] done, %d seconds\n",stime,done-start);
		fflush(ccfp);
		lock_unlock(fileno(ccfp));
		sv1log("%sCC [%d] done, %d requests / %d seconds\n",
			DST_PROTO,ccc,RequestSerno-startSerno,done-start);

		if( rcode != 0 ){
			sv1log("serverCC: exit by exit code=%d\n",rcode);
			break;
		}
	}
	close(svsock);
	if( ccfp != NULL )
	fclose(ccfp);
	unlink(path);
	sayFinishCC();
}

int connectToCache(Connection *Conn,PCStr(user),int *svsockp)
{	CStr(path,1024);
	int clsock,svport;

	clsock = tryOpen(Conn,user,0,AVStr(path),&svport);
	if( 0 <= clsock ){
		*svsockp = clsock;
		ServViaCc = 1;
		return 1;
	}else	return 0;
}

void set_CC()
{
	if( CC_Report[0] < 0 ){
		CC_master = getpid();
		Socketpair(CC_Report);
		setNonblockingIO(CC_Report[1],1);
		setCloseOnExec(CC_Report[0]);
		setCloseOnExec(CC_Report[1]);
	}
}
void putCCinfo(int pid)
{	int fd;

	if( (fd = CC_Report[1]) < 0 )
		return;
	write(fd,&pid,sizeof(pid));
}
void sayStartCC(){
	/* When the CC_master is not the parent of this process,
	 * it will not able to count down this CC by wait().
	 * Such CC will be fork()ed as a NNTPCC on timeout...
	 */
	if( getppid() != CC_master )
		return;
	putCCinfo(getpid());
}
void sayFinishCC(){
	if( getppid() != CC_master )
		return;
	putCCinfo(-getpid());
}
int num_CC()
{	int fd,pid;

	if( (fd = CC_Report[0]) < 0 )
		return 0;

	if( getpid() != CC_master )
		return 0;

	while( readTimeoutBlocked(fd,GVStr(&pid),sizeof(pid),1)==sizeof(pid) ){
		sv1log("## got_CCinfo: %d\n",pid);
		if( 0 <= pid ){
			sv1log("## add_CC [%d] %d\n",CC_Active,pid);
			CC_Procs[CC_Active++] = pid;
		}else{
			del_CC(-pid,"byRead");
		}
	}
	return CC_Active;
}
void del_CC(int pid,PCStr(how))
{	int ci;

	for( ci = 0; ci < CC_Active; ci++ ){
		if( CC_Procs[ci] == pid ){
			sv1log("## del_CC [%d/%d] %d %s\n",
				ci,CC_Active,pid,how);
			for(; ci < CC_Active; ci++ )
				CC_Procs[ci] = CC_Procs[ci+1];
			CC_Active--;
			break;
		}
	}
}
int kill_CC()
{	int ci,pid,nkill;

	nkill = 0;
	for( ci = 0; ci < CC_Active; ci++ ){
		if( 0 < (pid = CC_Procs[ci]) ){
			if( Kill(pid,SIGTERM) == 0 ){
				sv1log("Killed CC [pid=%d]\n",pid);
				nkill++;
			}
		}
	}
	return nkill;
}
